#include "tiams.h"
#include "Blocks.h"
#include "Blocksr1.h"


/*
Funktionprototypes
*/

static void AP_app (pFrame self, PEvent e);

int game_won (game_state *game_ptr);
int game_over (game_state *game_ptr);

int move (game_state *game_ptr, int direction);
void rotate (game_state *game_ptr, int rotation_number);
int move_up (game_state *game_ptr);

int find_pairs_vertically (game_state *game_ptr);

void check_and_add_to_highsore_list (game_state *game_ptr);
int can_be_new_highscore_entry (game_state *game_ptr);
void add_to_highscore_list (game_state *game_ptr, highscore *new_highscore_entry);
void show_highscore_dialog (game_state *game_ptr);

void draw_game (game_state *game_ptr);
void draw_update_game (game_state *old_game_ptr, game_state *new_game_ptr);
void draw_block (int i, int j, long value);

void new_game (game_state *game_ptr);
void create_random_tile (game_state *game_ptr);
void create_new_tile (game_state *game_ptr, long value);
block_pos find_random_empty_cell (game_state *game_ptr);

void set_null (game_state *game_ptr);
void set_merge_info_null (game_state *game_ptr);

void show_options_dialog (game_state *game_ptr);
void show_about_dialog (void);
void save_or_restore_highscore (game_state *game_ptr);

int random (int low, int high);

DWORD no_call_back (WORD DlgId, DWORD Value);


/*
Deklaration of the icon
*/

extern const MYICON app_icon;


/*
App object frame
*/

FRAME(appObj, OO_SYSTEM_FRAME, 0, OO_APP_FLAGS, 5)
	ATTR(OO_APP_FLAGS, APP_INTERACTIVE)
	ATTR(OO_APP_NAME, "Blocks on the Rocks")
	ATTR(OO_APP_PROCESS_EVENT, &AP_app)
	ATTR(OO_APP_DEFAULT_MENU, &app_menu)
	ATTR(OO_APP_ICON, &app_icon)
ENDFRAME


/*
Global variables
*/

pFrame pAppObj = (pFrame)&appObj;										/* Must be 1st! */
WINDOW appW;
const int game_version = 17;											/* Version 1.6 */
const int leet_delta[10] = {-13, -1, 12, -7, 3, 0, 2, -9, 10, -111};	/* 3A5T3R 3GG */
int leet_i;
char leet_char;
game_state current_game;
game_state old_game;


/*
Icon of the app
*/

const MYICON app_icon = {22, 22, {
0x00,0x00,0x00,0xE8,0xED,0x5C,0xA8,0xA9,0x50,0xC8,0xA9,0x9C,0xA8,0xA9,0x84,0xA8,
0xA9,0x44,0xEE,0xED,0x5C,0x00,0x00,0x00,0x00,0xFC,0x00,0x03,0x87,0x00,0x0E,0x01,
0xC0,0x18,0x00,0x60,0x30,0x00,0x30,0x60,0x00,0x18,0xE0,0x00,0x1C,0x80,0x00,0x04,
0xBD,0xDA,0xB4,0xA5,0x52,0xA4,0x39,0x53,0x30,0x25,0x53,0x10,0x25,0x52,0x90,0x25,
0xDA,0xB0
}};


/*
Images of the blocks
*/

const MYBITMAP block_0 = {18, 18, {
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00
}};

const MYBITMAP block_2 = {18, 18, {
0x00,0x00,0x00,0x00,0x00,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,0x00,0x03,0x00,0x00,
0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,0x30,0x00,
0x00,0x30,0x00,0x00,0x30,0x00,0x00,0x30,0x00,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,
0x00,0x00,0x00,0x00,0x00,0x00
}};

const MYBITMAP block_4 = {18, 18, {
0x00,0x00,0x00,0x00,0x00,0x00,0x30,0x00,0x00,0x30,0x00,0x00,0x30,0x00,0x00,0x30,
0x00,0x00,0x30,0xC0,0x00,0x30,0xC0,0x00,0x3F,0xFE,0x00,0x3F,0xFE,0x00,0x00,0xC0,
0x00,0x00,0xC0,0x00,0x00,0xC0,0x00,0x00,0xC0,0x00,0x00,0xC0,0x00,0x00,0xC0,0x00,
0x00,0x00,0x00,0x00,0x00,0x00
}};

const MYBITMAP block_8 = {18, 18, {
0x00,0x00,0x00,0x00,0x00,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,0x30,0x03,0x00,0x30,
0x03,0x00,0x30,0x03,0x00,0x30,0x03,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,0x30,0x03,
0x00,0x30,0x03,0x00,0x30,0x03,0x00,0x30,0x03,0x00,0x3F,0xFF,0x00,0x3F,0xFF,0x00,
0x00,0x00,0x00,0x00,0x00,0x00
}};

const MYBITMAP block_16 = {18, 18, {
0x00,0x00,0x00,0x00,0x00,0x00,0x3E,0x7F,0x00,0x3E,0x7F,0x00,0x06,0x60,0x00,0x06,
0x60,0x00,0x06,0x60,0x00,0x06,0x60,0x00,0x06,0x7F,0x00,0x06,0x7F,0x00,0x06,0x63,
0x00,0x06,0x63,0x00,0x06,0x63,0x00,0x06,0x63,0x00,0x06,0x7F,0x00,0x06,0x7F,0x00,
0x00,0x00,0x00,0x00,0x00,0x00
}};

const MYBITMAP block_32 = {18, 18, {
0xAA,0xAA,0x80,0x00,0x00,0x40,0xBF,0x3F,0x00,0x3F,0x3F,0x40,0x83,0x03,0x00,0x53,
0x53,0x40,0xAB,0x2B,0x00,0x43,0x03,0x40,0x9F,0x3F,0x00,0x5F,0x3F,0x40,0x83,0x30,
0x00,0x53,0x35,0x40,0xAB,0x32,0x80,0x03,0x30,0x40,0xBF,0x3F,0x00,0x3F,0x3F,0x40,
0x80,0x00,0x00,0x55,0x55,0x40
}};

const MYBITMAP block_64 = {18, 18, {
0xAA,0xAA,0x80,0x00,0x05,0x40,0xBF,0x32,0x80,0x3F,0x35,0x40,0xB0,0x32,0x80,0x35,
0x35,0x40,0xB2,0xB2,0x80,0x30,0x30,0x40,0xBF,0x36,0x00,0x3F,0x3F,0x40,0xB3,0x3F,
0x00,0x33,0x06,0x40,0xB3,0x26,0x80,0x33,0x56,0x40,0xBF,0x26,0x80,0x3F,0x56,0x40,
0x80,0x20,0x80,0x55,0x55,0x40
}};

const MYBITMAP block_128 = {18, 18, {
0xAA,0xAA,0x80,0x00,0x00,0x40,0xB7,0xDF,0x00,0x37,0xDF,0x40,0xB0,0xDB,0x00,0x34,
0xDB,0x40,0xB2,0xDB,0x00,0x30,0xDB,0x40,0xB7,0xDF,0x00,0x37,0xDF,0x40,0xB6,0x1B,
0x00,0x36,0x5B,0x40,0xB6,0x9B,0x00,0x36,0x1B,0x40,0xB7,0xDF,0x00,0x37,0xDF,0x40,
0x80,0x00,0x00,0x55,0x55,0x40
}};

const MYBITMAP block_256 = {18, 18, {
0xAA,0xAA,0x80,0x00,0x00,0x40,0xBD,0xEF,0x00,0x05,0x08,0x40,0xA5,0x2A,0x80,0x55,
0x49,0x40,0xA5,0x2A,0x80,0x05,0x08,0x40,0xBD,0xEF,0x00,0x20,0x29,0x40,0xAA,0xA9,
0x00,0x25,0x29,0x40,0xAA,0xA9,0x00,0x25,0x29,0x40,0xA0,0x29,0x00,0x3D,0xEF,0x40,
0x80,0x00,0x00,0x55,0x55,0x40
}};

const MYBITMAP block_512 = {18, 18, {
0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0xC3,0x10,0xC0,0xDF,0xDE,0xC0,0xDF,0xDE,0xC0,0xDF,
0xDE,0xC0,0xDF,0xDE,0xC0,0xDF,0xDE,0xC0,0xC3,0xD0,0xC0,0xFB,0xD7,0xC0,0xFB,0xD7,
0xC0,0xFB,0xD7,0xC0,0xFB,0xD7,0xC0,0xFB,0xD7,0xC0,0xFB,0xD7,0xC0,0xC3,0xD0,0xC0,
0xFF,0xFF,0xC0,0xFF,0xFF,0xC0
}};

const MYBITMAP block_1024 = {18, 18, {
0xFF,0xFF,0xC0,0x88,0x8B,0xC0,0xEA,0xEB,0xC0,0xEA,0xEB,0xC0,0xEA,0xEB,0xC0,0xEA,
0xEB,0xC0,0xEA,0xEB,0xC0,0xEA,0xEA,0xC0,0xEA,0x88,0x40,0xEA,0xBE,0xC0,0xEA,0xBE,
0xC0,0xEA,0xBE,0xC0,0xEA,0xBE,0xC0,0xEA,0xBE,0xC0,0xEA,0xBE,0xC0,0xEA,0xBE,0xC0,
0xE8,0x8E,0xC0,0xFF,0xFF,0xC0
}};

const MYBITMAP block_2048 = {18, 18, {
0xFF,0xFF,0xC0,0x80,0xC0,0x40,0xFE,0xDF,0x40,0xFE,0xDF,0x40,0x80,0xDF,0x40,0xBF,
0xDF,0x40,0xBF,0xDF,0x40,0x80,0xC0,0x40,0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0xBF,0xC0,
0x40,0xBF,0xDF,0x40,0xB7,0xDF,0x40,0x80,0xC0,0x40,0xF7,0xDF,0x40,0xF7,0xDF,0x40,
0xF7,0xC0,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_4096 = {18, 18, {
0xFF,0xFF,0xC0,0xBC,0x44,0x40,0xBD,0x55,0xC0,0xBD,0x55,0xC0,0xBD,0x55,0xC0,0xBD,
0x55,0xC0,0xBD,0x55,0xC0,0xAD,0x55,0xC0,0x85,0x45,0xC0,0xED,0x74,0x40,0xED,0x75,
0x40,0xED,0x75,0x40,0xED,0x75,0x40,0xED,0x75,0x40,0xED,0x75,0x40,0xED,0x75,0x40,
0xEC,0x44,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_8192 = {18, 18, {
0xFF,0xFF,0xC0,0x84,0x44,0x40,0xB7,0x57,0x40,0xB7,0x57,0x40,0xB7,0x57,0x40,0xB7,
0x57,0x40,0xB7,0x57,0x40,0xB7,0x57,0x40,0x87,0x44,0x40,0xB7,0x75,0xC0,0xB7,0x75,
0xC0,0xB7,0x75,0xC0,0xB7,0x75,0xC0,0xB7,0x75,0xC0,0xB7,0x75,0xC0,0xB7,0x75,0xC0,
0x87,0x44,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_16384 = {18, 18, {
0xFF,0xFF,0xC0,0x84,0x1F,0xC0,0xF5,0xFF,0xC0,0xF5,0xFF,0xC0,0xF4,0x1F,0xC0,0xF5,
0xDF,0xC0,0xF5,0xDF,0xC0,0xF4,0x17,0xC0,0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0x84,0x17,
0xC0,0xF5,0xD7,0xC0,0xF5,0xD5,0xC0,0xC4,0x10,0x40,0xF5,0xDD,0xC0,0xF5,0xDD,0xC0,
0x84,0x1D,0xC0,0xFF,0xFF,0xC0
}};

const MYBITMAP block_32768 = {18, 18, {
0xFF,0xFF,0xC0,0x84,0x1F,0xC0,0xF7,0xDF,0xC0,0xF7,0xDF,0xC0,0xC4,0x1F,0xC0,0xF5,
0xFF,0xC0,0xF5,0xFF,0xC0,0x84,0x17,0xC0,0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0x84,0x10,
0x40,0xF5,0xF7,0x40,0xF5,0xF7,0x40,0xF4,0x10,0x40,0xF5,0xD7,0x40,0xF5,0xD7,0x40,
0xF4,0x10,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_65536 = {18, 18, {
0xFF,0xFF,0xC0,0x82,0x1F,0xC0,0xBE,0xFF,0xC0,0xBE,0xFF,0xC0,0x82,0x1F,0xC0,0xBB,
0xDF,0xC0,0xBB,0xDF,0xC0,0x82,0x17,0xC0,0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0x82,0x10,
0x40,0xBF,0xD7,0xC0,0xBF,0xD7,0xC0,0x83,0x10,0x40,0xFB,0xD7,0x40,0xFB,0xD7,0x40,
0x82,0x10,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_131072 = {18, 18, {
0xFF,0xFF,0xC0,0x86,0x10,0xC0,0xF7,0xDE,0xC0,0xF7,0xDE,0xC0,0xF7,0x1E,0xC0,0xF7,
0xDE,0xC0,0xF7,0xDE,0xC0,0xF6,0x1E,0xC0,0xFF,0xFF,0xC0,0xFF,0xFF,0xC0,0x82,0x10,
0x40,0xBA,0xDF,0x40,0xBB,0xDF,0x40,0xBB,0xD0,0x40,0xBB,0xD7,0xC0,0xBB,0xD7,0xC0,
0x83,0xD0,0x40,0xFF,0xFF,0xC0
}};

const MYBITMAP block_unicorn = {18, 18, {
0x00,0x03,0x00,0x00,0x07,0x00,0x00,0xCE,0x00,0x03,0xFC,0x00,0x0E,0x38,0x00,0x18,
0x18,0x00,0x20,0x6C,0x00,0x60,0x66,0x00,0xC0,0x02,0x00,0x80,0x03,0x80,0x01,0x00,
0x80,0x01,0x80,0x80,0x01,0xC2,0xC0,0x01,0x7B,0x40,0x03,0x1C,0xC0,0x06,0x07,0x80,
0x06,0x00,0x00,0x04,0x00,0x00
}};


/*
Main function, called by the OS with the event as a parameter
*/

static void AP_app (pFrame self, PEvent e) {
	Access_AMS_Global_Variables;
	WIN_RECT appWR;
	short key;
	WORD opts;
	WORD fullscreen_flag;
	HANDLE pack_storage;	/* Handle to save and restore the current game state when the FlashApp is moved */

	switch (e->command) {
		case CM_INSTALL:
			set_null (&current_game);
			new_game (&current_game);
			old_game = current_game;
			break;
		case CM_DEFAULTS:
			set_null (&current_game);
			new_game (&current_game);
			old_game = current_game;
			break;
		case CM_PACK:
			pack_storage = HeapAlloc (sizeof (game_state) * 2);	/* Allocate memory for current_game and old_game */

			if (pack_storage != H_NULL) {
				game_state *pack_game;
				pack_game = HeapDeref (pack_storage);
				*pack_game = current_game;
				pack_game++;
				*pack_game = old_game;
			}

			OO_appSetPublicStorage (pack_storage);
			break;
		case CM_UNPACK:
			pack_storage = OO_appGetPublicStorage ();

			if (pack_storage != H_NULL) {
				game_state *pack_game;
				pack_game = HeapDeref (pack_storage);
				current_game = *pack_game;
				pack_game++;
				old_game = *pack_game;
				HeapFree (pack_storage);
			} else {
				set_null (&current_game);
				new_game (&current_game);
				old_game = current_game;
			}
			break;
		case CM_START:
			appWR = *(e->info.startInfo.startRect);

			MO_currentOptions ();
			if (MO_option[MO_OPT_SPLIT_SCREEN] == D_MODE_SPLIT_FULL) {
				fullscreen_flag = WF_NOBORDER;
			} else {
				fullscreen_flag = 0;
			}

			if (WinOpen (&appW, &appWR, WF_DUP_SCR | fullscreen_flag)) {
				WinAttr (&appW, A_REPLACE);
				WinClr (&appW);
			} else {
				EV_quit ();
			}
			break;
		case CM_ACTIVATE:
			EV_defaultHandler (e);
			WinBeginPaint (&appW);
			WinActivate (&appW);
			draw_game (&current_game);

			/* 3A5T3R 3GG */
			leet_i = 0;
			leet_char = 114;
			break;
		case CM_DEACTIVATE:
			WinEndPaint (&appW);
			EV_defaultHandler (e);
			break;
		case CM_QUIT:
			if (appW.Next) {
				WinClose (&appW);
				appW.Next = NULL;
			}
			break;
		case CM_KEY_PRESS:
			key = e->info.keyInfo.keyCode;
			if ((current_game.game_over == 0) && (key == KB_UP || key == KB_DOWN || key == KB_LEFT || key == KB_RIGHT)) {
				if (move (&current_game, key)) {
					draw_update_game (&old_game, &current_game);
					old_game = current_game;
				}

				set_merge_info_null (&current_game);	/* Reset the merge info */

				if (current_game.won == 0) {
					current_game.won = game_won (&current_game);
				}

				if (current_game.game_over == 0) {
					current_game.game_over = game_over (&current_game);
				}

				if (current_game.won == 1 && current_game.keep_playing == 0 && current_game.game_over == 0) {
					opts = 1;
					if (KB_ENTER == Dialog (&dialog_win_notification, -1, -1, NULL, &opts) && opts == 1) {
						current_game.keep_playing = 1;
					} else {
						current_game.game_over = 1;
						check_and_add_to_highsore_list (&current_game);
					}
				} else if (current_game.won == 1 && current_game.keep_playing == 0 && current_game.game_over == 1) {	/* User got a 2048 tile but no further moves are possible */
					opts = 1;
					Dialog (&dialog_win_and_game_over, -1, -1, NULL, NULL);
					check_and_add_to_highsore_list (&current_game);
				} else if ( ((current_game.won == 1 && current_game.keep_playing == 1) || (current_game.won == 0 && current_game.keep_playing == 0)) && current_game.game_over == 1) {
					opts = 1;
					Dialog (&dialog_game_over, -1, -1, NULL, NULL);
					check_and_add_to_highsore_list (&current_game);
				}

			} else {
				/* 3A5T3R 3GG */
				if (key == leet_char) {
					leet_char += leet_delta[leet_i];
					leet_i++;
					if (!leet_char) {
						leet_char = 114;
						leet_i = 0;
						save_or_restore_highscore (&current_game);
					}
				} else if (key == 114) {
					leet_i = 1;
					leet_char = 114 + leet_delta[0];
				} else {
					leet_char = 114;
					leet_i = 0;
					EV_defaultHandler (e);
				}
			}
			break;
		case CM_WPAINT:
			DrawWinBorder (&appW, &appW.Window);
			WinBackupToScr (&appW);
			break;
		case ACM_NEW:
			WinClr (&appW);
			new_game (&current_game);
			draw_game (&current_game);
			old_game = current_game;
			break;
		case ACM_HIGHSCORE:
			show_highscore_dialog (&current_game);
			break;
		case ACM_OPTIONS:
			show_options_dialog (&current_game);
			break;
		case ACM_ABOUT:
			show_about_dialog ();
			break;
		default:
			EV_defaultHandler (e);
			break;
	}
}


/*
Check if game is won (check for the magical 2048 tile)
*/

int game_won (game_state *game_ptr) {
	int won;
	int i;
	int j;

	won = 0;
	i = 0;
	while (i <= 3) {
		j = 0;
		while (j <= 3) {
			if (game_ptr->grid[i][j] == 2048) {
				won = 1;
			}
			j++;
		}
		i++;
	}

	return won;
}


/*
Check if game is over/lost
Returns 1 if game is lost and 0 if game is not lost
*/

int game_over (game_state *game_ptr) {
	int result;

	result = 1;

	if (find_pairs_vertically (game_ptr) == 1 || game_ptr->empty_cells > 0) {
		return 0;
	}

	rotate (game_ptr, 1);
	if (find_pairs_vertically (game_ptr) == 1) {
		result = 0;
	}
	rotate (game_ptr, 3);

	return result;
}


/*
Move tiles in the specified direction and join tiles if possible
*/

int move (game_state *game_ptr, int direction) {
	int changed;

	switch (direction) {
		case KB_UP:
			changed = move_up (game_ptr);
			break;
		case KB_DOWN:
			rotate (game_ptr, 2);
			changed = move_up (game_ptr);
			rotate (game_ptr, 2);
			break;
		case KB_LEFT:
			rotate (game_ptr, 1);
			changed = move_up (game_ptr);
			rotate (game_ptr, 3);
			break;
		case KB_RIGHT:
			rotate (game_ptr, 3);
			changed = move_up (game_ptr);
			rotate (game_ptr, 1);
			break;
		default:
			changed = 0;
			break;
	}

	return changed;
}


/*
Rotate the grid clockwise, rotation_number specifies how many times the grid is rotated
*/

void rotate (game_state *game_ptr, int rotation_number) {
	int i;
	int j;
	int tmp;

	if (rotation_number < 0) {
		rotation_number = 4 - ((-rotation_number) % 4);
	} else {
		rotation_number = rotation_number % 4;
	}

	if (rotation_number != 0) {
		while (rotation_number-- != 0) {
			i = 0;
			while (i < game_ptr->current_options.size / 2) {
				j = i;
				while (j < game_ptr->current_options.size - i - 1) {
					tmp = game_ptr->grid[i][j];
					game_ptr->grid[i][j] = game_ptr->grid[j][game_ptr->current_options.size - i - 1];
					game_ptr->grid[j][game_ptr->current_options.size - i - 1] = game_ptr->grid[game_ptr->current_options.size - i - 1][game_ptr->current_options.size - j - 1];
					game_ptr->grid[game_ptr->current_options.size - i - 1][game_ptr->current_options.size - j - 1] = game_ptr->grid[game_ptr->current_options.size - j - 1][i];
					game_ptr->grid[game_ptr->current_options.size - j - 1][i] = tmp;
					j++;
				}
				i++;
			}
		}
	}
}


/*
Move tiles up and join tiles if possible
*/

int move_up (game_state *game_ptr) {
	int changed;
	int i;
	int j;
	int k;

	changed = 0;

	j = 1;
	while (j < game_ptr->current_options.size) {
		i = 0;
		while (i < game_ptr->current_options.size) {
			if (game_ptr->grid[i][j] != 0) {
				k = j - 1;
				while (1) {
					if (k < 0) {
						k++;
						game_ptr->grid[i][k] = game_ptr->grid[i][j];
						game_ptr->grid[i][j] = 0;
						changed = 1;
						break;
					}
					if (game_ptr->grid[i][k] != 0) {
						if (game_ptr->grid[i][k] == game_ptr->grid[i][j] && game_ptr->grid_merged[i][k] == 0) {
							game_ptr->grid[i][k] *= 2;
							game_ptr->grid[i][j] = 0;
							game_ptr->score += game_ptr->grid[i][k];
							game_ptr->grid_merged[i][k] = 1;
							game_ptr->empty_cells++;
							changed = 1;
						} else if (j - k != 1) {
							k++;
							game_ptr->grid[i][k] = game_ptr->grid[i][j];
							game_ptr->grid[i][j] = 0;
							changed = 1;
						}
						break;
					}
					k--;
				}
			}
			i++;
		}
		j++;
	}

	if (game_ptr->best_score[game_ptr->current_options.config_id] < game_ptr->score) {
		game_ptr->best_score[game_ptr->current_options.config_id] = game_ptr->score;
	}

	if (changed == 1) {
		create_random_tile (game_ptr);
		game_ptr->moves++;
	}

	return changed;
}


/*
Check if there are vertical pairs in the grid
Returns 1 if a pair was found and 0 if no pair was found
*/

int find_pairs_vertically (game_state *game_ptr) {
	int i;
	int j;

	i = 0;
	while (i < game_ptr->current_options.size) {
		j = 0;
		while (j < game_ptr->current_options.size - 1) {
			if (game_ptr->grid[i][j] == game_ptr->grid[i][j + 1]) {
				return 1;
			}
			j++;
		}
		i++;
	}

	return 0;
}


/*
If the score is high enough to get into the highscore list, ask for the user's name and add the new score to the highscore
*/

void check_and_add_to_highsore_list (game_state *game_ptr) {
	highscore new_highscore_entry;

	if (can_be_new_highscore_entry (game_ptr) == 1) {
		new_highscore_entry.score = game_ptr->score;

		strcpy (new_highscore_entry.name, "");
		if (KB_ENTER == Dialog (&dialog_new_highscore, -1, -1, new_highscore_entry.name, NULL)) {
			if (new_highscore_entry.name[0] == '\0') {	/* empty name string -> "no name" */
				strcpy (new_highscore_entry.name, "no name");
			}
			add_to_highscore_list (game_ptr, &new_highscore_entry);
		}
	}
}


/*
Check if the score is high enough to get into the highscore list
*/

int can_be_new_highscore_entry (game_state *game_ptr) {
	if (game_ptr->highscore_list[game_ptr->current_options.config_id][4].score < game_ptr->score) {
		return 1;
	} else {
		return 0;
	}
}


/*
Add an highscore entry to the highscore list
*/

void add_to_highscore_list (game_state *game_ptr, highscore *new_highscore_entry) {
	int i;
	int j;

	i = 4;
	while (i >= 0) {
		if (game_ptr->highscore_list[game_ptr->current_options.config_id][i].score >= new_highscore_entry->score) {
			break;
		}
		i--;
	}
	i++;

	if (i < 5) {
		j = 3;
		while (j >= i) {
			game_ptr->highscore_list[game_ptr->current_options.config_id][j + 1] = game_ptr->highscore_list[game_ptr->current_options.config_id][j];
			j--;
		}

		game_ptr->highscore_list[game_ptr->current_options.config_id][i] = *new_highscore_entry;
	}
}


/*
Show a dialog with the highscore list
*/

void show_highscore_dialog (game_state *game_ptr) {
	char highscore_string[256];

	sprintf (highscore_string, "1. %-11s %6li\n2. %-11s %6li\n3. %-11s %6li\n4. %-11s %6li\n5. %-11s %6li",
		game_ptr->highscore_list[game_ptr->current_options.config_id][0].name, game_ptr->highscore_list[game_ptr->current_options.config_id][0].score,
		game_ptr->highscore_list[game_ptr->current_options.config_id][1].name, game_ptr->highscore_list[game_ptr->current_options.config_id][1].score,
		game_ptr->highscore_list[game_ptr->current_options.config_id][2].name, game_ptr->highscore_list[game_ptr->current_options.config_id][2].score,
		game_ptr->highscore_list[game_ptr->current_options.config_id][3].name, game_ptr->highscore_list[game_ptr->current_options.config_id][3].score,
		game_ptr->highscore_list[game_ptr->current_options.config_id][4].name, game_ptr->highscore_list[game_ptr->current_options.config_id][4].score);
	DlgNotice ("Highscore", highscore_string);
}


/*
Draw the current gamestate to the main window
*/

void draw_game (game_state *game_ptr) {
	WIN_RECT line;
	char score_str[10];
	int i;
	int j;

	line.x0 = 10;
	line.y0 = 10;
	line.x1 = game_ptr->current_options.size * 19 + 10;
	line.y1 = 10;

	i = 1;
	while (i <= game_ptr->current_options.size + 1) {
		WinLine (&appW, &line);
		line.y0 += 19;
		line.y1 += 19;
		i++;
	}

	line.x0 = 10;
	line.y0 = 10;
	line.x1 = 10;
	line.y1 = game_ptr->current_options.size * 19 + 10;

	i = 1;
	while (i <= game_ptr->current_options.size + 1) {
		WinLine (&appW, &line);
		line.x0 += 19;
		line.x1 += 19;
		i++;
	}

	i = 0;
	while (i < game_ptr->current_options.size) {
		j = 0;
		while (j < game_ptr->current_options.size) {
			if (game_ptr->grid[i][j] != 0) {
				draw_block (i, j, game_ptr->grid[i][j]);
			}
			j++;
		}
		i++;
	}

	sprintf (score_str, "%li", game_ptr->score);
	WinStrXY (&appW, 106, 17, "Blocks on the Rocks");
	WinStrXY (&appW, 120, 38, "Score: ");
	WinStr (&appW, score_str);
	WinStrXY (&appW, 120, 58, "Best : ");
	if (game_ptr->best_score[game_ptr->current_options.config_id] >= game_ptr->score) {
		sprintf (score_str, "%li", game_ptr->best_score[game_ptr->current_options.config_id]);
		WinStr (&appW, score_str);
	} else {
		WinStr (&appW, score_str);
	}
}


/*
Update the game window, only the things that have changed are redrawn
The function assumes that the score is increasing
*/

void draw_update_game (game_state *old_game_ptr, game_state *new_game_ptr) {
	char score_str[10];
	int i;
	int j;

	i = 0;
	while (i < new_game_ptr->current_options.size) {
		j = 0;
		while (j < new_game_ptr->current_options.size) {
			if (old_game_ptr->grid[i][j] != new_game_ptr->grid[i][j]) {	/* Check if tile needs to be redrawn */
				draw_block (i, j, new_game_ptr->grid[i][j]);
			}
			j++;
		}
		i++;
	}

	sprintf (score_str, "%li", new_game_ptr->score);
	WinStrXY (&appW, 162, 38, score_str);
	if (new_game_ptr->best_score[new_game_ptr->current_options.config_id] >= new_game_ptr->score) {
		sprintf (score_str, "%li", new_game_ptr->best_score[new_game_ptr->current_options.config_id]);
		WinStrXY (&appW, 162, 58, score_str);
	} else {
		WinStrXY (&appW, 162, 58, score_str);
	}
}


/*
Draw a single block
*/

void draw_block (int i, int j, long value) {
	switch (value) {
		case -1:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_unicorn, A_REPLACE);
			break;
		case 0:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_0, A_REPLACE);
			break;
		case 2:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_2, A_REPLACE);
			break;
		case 4:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_4, A_REPLACE);
			break;
		case 8:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_8, A_REPLACE);
			break;
		case 16:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_16, A_REPLACE);
			break;
		case 32:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_32, A_REPLACE);
			break;
		case 64:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_64, A_REPLACE);
			break;
		case 128:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_128, A_REPLACE);
			break;
		case 256:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_256, A_REPLACE);
			break;
		case 512:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_512, A_REPLACE);
			break;
		case 1024:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_1024, A_REPLACE);
			break;
		case 2048:			/* The magical 2048 tile! */
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_2048, A_REPLACE);
			break;
		case 4096:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_4096, A_REPLACE);
			break;
		case 8192:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_8192, A_REPLACE);
			break;
		case 16384:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_16384, A_REPLACE);
			break;
		case 32768:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_32768, A_REPLACE);
			break;
		case 65536:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_65536, A_REPLACE);
			break;
		case 131072:
			WinBitmapPut (&appW, i * 19 + 11, j * 19 + 11, (BITMAP *)&block_131072, A_REPLACE);
			break;
	}
}


/*
Reset the current game and start a new game with the new game mode
*/

void new_game (game_state *game_ptr) {
	int i;
	int j;

	game_ptr->current_options = game_ptr->new_options;

	i = 0;									/* Reset the whole grid */
	while (i < game_ptr->current_options.size) {
		j = 0;
		while (j < game_ptr->current_options.size) {
			game_ptr->grid[i][j] = 0;
			j++;
		}
		i++;
	}

	game_ptr->score = 0;					/* Reset the score and     */
	game_ptr->moves = 0;					/* the number of moves     */
	game_ptr->keep_playing = 0;				/* and the gamestate flags */
	game_ptr->game_over = 0;
	game_ptr->won = 0;
	game_ptr->empty_cells = game_ptr->current_options.size * game_ptr->current_options.size;

	create_random_tile (game_ptr);			/* Two new tiles are created */
	create_random_tile (game_ptr);

	if (game_ptr->new_options.mode == MODE_UNICORN) {
		create_new_tile (game_ptr, -1);
	}

}


/*
Create a random tile
*/

void create_random_tile (game_state *game_ptr) {
	int rand_val;
	long new_block_value;

	rand_val = random (1, 100);
	if (rand_val > 90) {					/* The chance for a 4-tile is 10% */
		new_block_value = 4;
	} else {								/* The chance for a 2-tile is 90% */
		new_block_value = 2;
	}

	create_new_tile (game_ptr, new_block_value);
}

/*
Create new tile
*/

void create_new_tile (game_state *game_ptr, long value) {
	block_pos new_block_pos;

	new_block_pos = find_random_empty_cell (game_ptr);
	game_ptr->grid[new_block_pos.x][new_block_pos.y] = value;
	game_ptr->empty_cells--;
}


/*
Find a random empty cell
*/

block_pos find_random_empty_cell (game_state *game_ptr) {
	block_pos random_position;
	int rand_val;
	int done;
	int i;
	int j;

	rand_val = random (1, game_ptr->empty_cells);
	done = 0;
	j = 0;
	i = 0;
	while (i < game_ptr->current_options.size) {
		j = 0;
		while (j < game_ptr->current_options.size) {
			if (game_ptr->grid[i][j] == 0) {
				rand_val--;
				if (rand_val == 0) {
					done = 1;
					break;
				}
			}
			j++;
		}
		if (done == 1) {
			break;
		}
		i++;
	}

	random_position.x = i;
	random_position.y = j;

	return random_position;
}


/*
Reset the merge info to zero
*/

void set_merge_info_null (game_state *game_ptr) {
	int i;
	int j;

	i = 0;
	while (i < game_ptr->current_options.size) {
		j = 0;
		while (j < game_ptr->current_options.size) {
			game_ptr->grid_merged[i][j] = 0;
			j++;
		}
		i++;
	}
}


/*
Initialize the program's main datastructure with zeros
*/

void set_null (game_state *game_ptr) {
	int i;
	int j;

	game_ptr->score = 0;

	i = 0;
	while (i <= 5) {
		game_ptr->best_score[i] = 0;
		i++;
	}
	game_ptr->moves = 0;
	game_ptr->won = 0;
	game_ptr->game_over = 0;
	game_ptr->keep_playing = 0;
	game_ptr->current_options.mode = MODE_CLASSIC;
	game_ptr->current_options.size = 4;
	game_ptr->current_options.animations = 0;
	game_ptr->current_options.undo = 0;
	game_ptr->current_options.config_id = ((game_ptr->current_options.mode - 1) * 3) + (game_ptr->current_options.size - 2);
	game_ptr->new_options.mode = MODE_CLASSIC;
	game_ptr->new_options.size = 4;
	game_ptr->new_options.animations = 0;
	game_ptr->new_options.undo = 0;
	game_ptr->new_options.config_id = ((game_ptr->new_options.mode - 1) * 3) + (game_ptr->new_options.size - 2);
	game_ptr->empty_cells = 0;

	i = 0;
	while (i <= 3) {
		j = 0;
		while (j <= 3) {
			game_ptr->grid[i][j] = 0;
			game_ptr->grid_merged[i][j] = 0;
			j++;
		}
		i++;
	}

	i = 0;
	while (i <= 5) {
		j = 0;
		while (j <= 4) {
			strcpy(game_ptr->highscore_list[i][j].name, "-");
			game_ptr->highscore_list[i][j].score = 0;
			j++;
		}
		i++;
	}
}


/*
Display the config dialog
*/

void show_options_dialog (game_state *game_ptr) {
	WORD options_input[2];

	options_input[0] = game_ptr->new_options.mode;
	options_input[1] = game_ptr->new_options.size;

	if (KB_ENTER == Dialog (&dialog_options, -1, -1, NULL, options_input)) {
		game_ptr->new_options.mode = options_input[0];
		game_ptr->new_options.size = options_input[1];
		game_ptr->new_options.config_id = ((game_ptr->new_options.mode - 1) * 3) + (game_ptr->new_options.size - 2);
	}
}


/*
Display the Credits dialog
*/

void show_about_dialog (void) {
	DlgNotice ("About", "Blocks on the Rocks V1.7\n\nAuthors:\n  Daniel Eisele\n  (daniel_eisele@gmx.de)\n  Janek Schoffit\n\nWebsite:\n  http://daniel.ibme.de");
}


/*
If the save file exists, restore the highscore from the save file.
If the save file does not exist create it and save the highscore to it.

Save file format:
  int game_version               - Version of the game which created the file
  highscore highscore_list[x][5] - Array of highscore entries
  long int best_score[x]         - Array of the best scores
*/

void save_or_restore_highscore (game_state *game_ptr) {
	FILES save_file;
	FSWORD file_size;
	int success;
	highscore highscore_read_buffer[6][5];
	long int best_score_read_buffer[6];
	int highscore_version;
	char file_name[] = "blockshs";	/* max. 8 characters */
	char file_type[] = "HIGH";		/* max. 4 characters */

	success = 0;

	if (FS_OK == FAccess (file_name, FM_READ_ACCESS, file_type)) {																					/* Check if file exists */
		if (FS_OK == FOpen (file_name, &save_file, FM_READ, file_type)) {																			/* Open file */
			if (FS_OK == FRead (&highscore_version, sizeof (highscore_version), &save_file)) {														/* Read the highscore version */
				if (highscore_version <= game_version) {																							/* Only highscore files from this version or a lower version can be read */
					if (sizeof (highscore_version) + sizeof (highscore_read_buffer) + sizeof (best_score_read_buffer) == FGetSize (&save_file)) {	/* Check file size */
						if (FS_OK == FRead (highscore_read_buffer, sizeof (highscore_read_buffer), &save_file)) {									/* Read highscore list into buffer */
							if (FS_OK == FRead (best_score_read_buffer, sizeof (best_score_read_buffer), &save_file)) {								/* Read best scores list int buffer */
								memcpy (game_ptr->highscore_list, highscore_read_buffer, sizeof (highscore_read_buffer));							/* Copy highscore list into the game_state datastructure */
								memcpy (game_ptr->best_score, best_score_read_buffer, sizeof (best_score_read_buffer));								/* Copy best scores list into the game_state datastructure */
								success = 1;
							}
						}
					}
				}
			}
			FClose (&save_file);		/* Close file again */
		}
	} else if (FS_OK == FAccess (file_name, FM_WRITE_ACCESS, file_type)) {
		if (FS_OK == FOpen (file_name, &save_file, FM_WRITE, file_type)) {
			if (FS_OK == FWrite (&game_version, sizeof (game_version), &save_file)) {
				if (FS_OK == FWrite (game_ptr->highscore_list, sizeof (game_ptr->highscore_list), &save_file)) {
					if (FS_OK == FWrite (game_ptr->best_score, sizeof (game_ptr->best_score), &save_file)) {
						success = 2;
					}
				}
			}
			FClose (&save_file);		/* Close file again */
		}
	}

	if (success == 1) {
		DlgNotice ("Success", "Highscore restored.");
	} else if (success == 2) {
		DlgNotice ("Success", "Highscore saved to blockshs.");
	} else {
		DlgNotice ("Failure", "Highscore could not be saved or restored.");
	}
}


/*
Pseudo random number generator
Taken from the pipes example program in the "TI-89/TI-92 Plus Developer Guide" on page 47/48
*/

int random (int low, int high) {
	int range;
	static long seed = 29;

	if (seed == 0) {
		seed = 1;
	}
	
	seed *= 16807;
	
	if (seed < 0) {
		seed = -seed;
	}
	
	range = high - low + 1;
	return low + seed % range;
}


/*
Callback function for dialogs
This function is provided as a callback function when a dialog is opened.
Because we don't need a callback function here, we just leave it empty.
*/

DWORD no_call_back (WORD DlgId, DWORD Value) {
	return TRUE;
} /* Take a look at the line count */
