#define ACM_NEW			0x500
#define ACM_HIGHSCORE	0x501
#define ACM_OPTIONS		0x502
#define ACM_ABOUT		0x503

#define SIZE			4

#define UP				1
#define DOWN 			2
#define LEFT			3
#define RIGHT			4

#define MAX_FRAMES		5

#define MODE_CLASSIC	1
#define MODE_UNICORN	2

typedef struct {
	int x;
	int y;
} block_pos;

typedef struct {					/* struct for one highscore entry */
	char name[12];
	long int score;
} highscore;

typedef struct {
	int mode;						/* can be MODE_*, default is MODE_CLASSIC */
	int size;						/* grid size (from 2 to 4), default is 4 */
	int animations;					/* 0: animations off, 1: animations on */
	int undo;						/* 0: undo is off, 1: undo is on*/
	int config_id;					/* Each combination of config options has its number which will be used when accessing the highscore */
} options;

typedef struct {
	int frame;						/* current animation frame number. -1 means that no animation is running */
	int max_frames;					/* maximal frame number */
} animation_state;

typedef struct {					/* struct for the complete program state */
	highscore highscore_list[6][5];	/* five highscore entries in each highscore list */
	long int score;					/* score in the current game */
	long int best_score[6];			/* best score (unrelated to the highscore list) */
	int moves;						/* number of moves in the current game */
	int won;
	int game_over;
	int keep_playing;
	options new_options;			/* new settings of the game. they will be applied when a new game is started */
	options current_options;		/* current settings of the game */
	animation_state animation;		/* State of the current animation*/
	long grid[4][4];				/* grid[x][y]; */
	int grid_merged[4][4];			/* Store if a tile was already merged to prevent merging a tile twice in a single turn */
	int empty_cells;				/* Number of empty cells in the grid */	
} game_state;

typedef struct {					/* struct for a bitmap with a predefined size */
	WORD NumRows;
	WORD NumCols;
	BYTE Data[54];
} MYBITMAP;

typedef struct {					/* struct for a icon with a predefined size */
	WORD NumRows;
	WORD NumCols;
	BYTE Data[66];
} MYICON;